#
# Copyright (c) 2021 Carsten Igel.
#
# This file is part of nt2dc
# (see https://github.com/carstencodes/nt2dc).
#
# License: 3-clause BSD, see https://opensource.org/licenses/BSD-3-Clause
#
"""Replacements for generic types built in python 3.8 or earlier.
"""

import typing
import collections
import collections.abc
import contextlib
import re

BuiltInReplacements: typing.Dict[typing.Type, typing.Type] = {
    tuple: typing.Tuple,
    collections.abc.Callable: typing.Callable,
    type: typing.Type,
    dict: typing.Dict,
    list: typing.List,
    set: typing.Set,
    frozenset: typing.FrozenSet,
    collections.defaultdict: typing.DefaultDict,
    collections.OrderedDict: typing.OrderedDict,
    collections.ChainMap: typing.ChainMap,
    collections.Counter: typing.Counter,
    collections.deque: typing.Deque,
    re.Pattern: typing.Pattern,
    re.Match: typing.Match,
    collections.abc.Set: typing.Set,
    collections.abc.ByteString: typing.ByteString,
    collections.abc.Collection: typing.Collection,
    collections.abc.Container: typing.Container,
    collections.abc.ItemsView: typing.ItemsView,
    collections.abc.KeysView: typing.KeysView,
    collections.abc.Mapping: typing.Mapping,
    collections.abc.MappingView: typing.MappingView,
    collections.abc.MutableMapping: typing.MutableMapping,
    collections.abc.MutableSequence: typing.MutableSequence,
    collections.abc.MutableSet: typing.MutableSet,
    collections.abc.Sequence: typing.Sequence,
    collections.abc.ValuesView: typing.ValuesView,
    collections.abc.Iterable: typing.Iterable,
    collections.abc.Iterator: typing.Iterator,
    collections.abc.Generator: typing.Generator,
    collections.abc.Reversible: typing.Reversible,
    collections.abc.Coroutine: typing.Coroutine,
    collections.abc.AsyncGenerator: typing.AsyncGenerator,
    collections.abc.AsyncIterable: typing.AsyncIterable,
    collections.abc.AsyncIterator: typing.AsyncIterator,
    collections.abc.Awaitable: typing.Awaitable,
    contextlib.AbstractContextManager: typing.ContextManager,
    contextlib.AbstractAsyncContextManager: typing.AsyncContextManager,
}

BuiltInNames: typing.Dict[typing.Type, str] = {
    typing.Tuple: "typing.Tuple",
    typing.Callable: "typing.Callable",
    typing.Type: "typing.Type",
    typing.Dict: "typing.Dict",
    typing.List: "typing.List",
    typing.FrozenSet: "typing.FrozenSet",
    typing.DefaultDict: "typing.DefaultDict",
    typing.OrderedDict: "typing.OrderedDict",
    typing.ChainMap: "typing.ChainMap",
    typing.Counter: "typing.Counter",
    typing.Deque: "typing.Deque",
    typing.Pattern: "typing.Pattern",
    typing.Match: "typing.Match",
    typing.Set: "typing.Set",
    typing.ByteString: "typing.ByteString",
    typing.Collection: "typing.Collection",
    typing.Container: "typing.Container",
    typing.ItemsView: "typing.ItemsView",
    typing.KeysView: "typing.KeysView",
    typing.Mapping: "typing.Mapping",
    typing.MappingView: "typing.MappingView",
    typing.MutableMapping: "typing.MutableMapping",
    typing.MutableSequence: "typing.MutableSequence",
    typing.MutableSet: "typing.MutableSet",
    typing.Sequence: "typing.Sequence",
    typing.ValuesView: "typing.ValuesView",
    typing.Iterable: "typing.Iterable",
    typing.Iterator: "typing.Iterator",
    typing.Generator: "typing.Generator",
    typing.Reversible: "typing.Reversible",
    typing.Coroutine: "typing.Coroutine",
    typing.AsyncGenerator: "typing.AsyncGenerator",
    typing.AsyncIterable: "typing.AsyncIterable",
    typing.AsyncIterator: "typing.AsyncIterator",
    typing.Awaitable: "typing.Awaitable",
    typing.ContextManager: "typing.ContextManager",
    typing.AsyncContextManager: "typing.AsyncContextManager",
}



#
# Copyright (c) 2021 Carsten Igel.
#
# This file is part of nt2dc
# (see https://github.com/carstencodes/nt2dc).
#
# License: 3-clause BSD, see https://opensource.org/licenses/BSD-3-Clause
#
""" Provides methods to convert a typed named tuple to a dataclass.
"""

from typing import (
    FrozenSet,
    Set,
    Mapping,
    Type,
    NamedTuple,
    Tuple,
    List,
    Any,
    Dict,
    get_type_hints,
    get_args,
    get_origin,
)
from dataclasses import make_dataclass as make_real_dataclass
from collections.abc import Iterable
from sys import version_info as python_version
from importlib import import_module

from .builtins import BuiltInReplacements, BuiltInNames


__BUILT_IN_REPLACEMENTS__ = None
if python_version.major == 3 and python_version.minor < 9:
    __BUILT_IN_REPLACEMENTS__ = BuiltInReplacements

__BUILT_IN_CONSTRUCTORS__ = {
    List: list,
    Dict: dict,
    Set: set,
    FrozenSet: frozenset,
    Tuple: tuple,
}


def make_dataclass(
    clz: Type[NamedTuple],
    generic_type_mapping: Mapping[Type, Type] = __BUILT_IN_REPLACEMENTS__,
) -> Type[object]:
    """Creates a dataclass from the specified named tuple.

    Args:
        clz (Type[NamedTuple]): The type to create a dataclass from.
        generic_type_mapping (Mapping[Type, Type], optional): An
            optional mapping of source types to target types, e.g. from set
            to list. Defaults to None.

    Returns:
        Type[object]: The dataclass.
    """
    type_hints: List[Tuple[str, Type]] = []
    for key, value in get_type_hints(clz).items():
        value = _get_target_data_type_dc(value, generic_type_mapping)
        type_hints.append((key, value))

    return make_real_dataclass(clz.__name__ + "DataClass", type_hints)


def _get_target_data_type_dc(
    value: Type, generic_type_mapping: Mapping[Type, Type] = None
) -> Type:
    args: Tuple = get_args(value)
    if len(args) > 0:
        items: List[Type] = []
        for arg in args:
            arg = _get_target_data_type_dc(arg)
            items.append(arg)

        origin: Type = get_origin(value)
        if (
            generic_type_mapping is not None
            and origin in generic_type_mapping.keys()
        ):
            origin = generic_type_mapping[origin]
        origin = _get_target_data_type_dc(origin)
        value = _make_generic_type(origin, items)
    elif _is_namedtuple_instance(value):
        value = make_dataclass(value)

    return value


def _make_generic_type(generic_base: Type, generic_args: List[Type]) -> Type:
    _globals: Dict[str, Any] = {}

    def qualname(queried_type) -> str:
        if hasattr(queried_type, "__module__"):
            module: Any = import_module(queried_type.__module__)
            _globals[queried_type.__module__] = module
        if queried_type in BuiltInNames.keys():
            return BuiltInNames[queried_type]

        if hasattr(queried_type, "__qualname__"):
            return queried_type.__qualname__

        return queried_type.__name__

    base_type_name: str = qualname(generic_base)
    generic_arg_names: List[str] = [qualname(t) for t in generic_args]
    generic_arg_names_value = ", ".join(generic_arg_names)
    target_type_qualified_name = "{}[{}]".format(
        base_type_name, generic_arg_names_value
    )
    # At this stage, there is no other way than using an eval statement
    # with a qualified name like list[int] or typing.Dict[str, str]
    # pylint: disable=W0123
    result: Type = eval(target_type_qualified_name, _globals)
    return result


def get_dataclass_object(
    instance: NamedTuple,
    generic_type_mapping: Mapping[Type, Type] = __BUILT_IN_REPLACEMENTS__,
) -> Tuple[object, Type[object]]:
    """Creates a dataclass object from the specified named tuple instance.

    Args:
        instance (NamedTuple): The instance to create a dataclass instance
            from.
        generic_type_mapping (Mapping[Type, Type], optional): An
            optional mapping of source types to target types, e.g. from set
            to list. Defaults to None.
    Returns:
        Tuple[object, Type[object]]: The converted instance and its new type.
    """
    target_type: Type[object] = make_dataclass(
        instance.__class__, generic_type_mapping
    )
    items_narrowed: Dict[str, Any] = _narrow_named_tuple_instance(
        instance, generic_type_mapping
    )
    new_instance: object = target_type(**items_narrowed)

    return (new_instance, target_type)


def _narrow_named_tuple_instance(
    instance: Any,
    generic_type_mapping: Mapping[Type, Type] = __BUILT_IN_REPLACEMENTS__,
) -> Dict[str, Any]:
    result: Dict[str, Any] = {}
    for key, value in instance._asdict().items():
        key_type: Type = type(value)
        if value is NamedTuple:
            value = _narrow_named_tuple_instance(value, generic_type_mapping)
        elif value is Iterable:
            items = {v is NamedTuple for v in value}
            if True in items:
                narrowed_items = []
                for item in value:
                    narrowed_items.append(
                        _narrow_named_tuple_instance(
                            item, generic_type_mapping
                        )
                    )

                value = narrowed_items

        if (
            generic_type_mapping is not None
            and key_type in generic_type_mapping.keys()
        ):
            target_type: Type = generic_type_mapping[key_type]
            ctor = target_type
            if target_type in __BUILT_IN_CONSTRUCTORS__.keys():
                ctor = __BUILT_IN_CONSTRUCTORS__[target_type]
            value = ctor(value)

        result[key] = value

    return result


def _is_namedtuple_instance(obj: Any) -> bool:
    is_tuple: bool = isinstance(obj, tuple)
    has_asdict_method: bool = hasattr(obj, "_asdict")
    has_fields_class_member: bool = hasattr(obj, "_fields")

    return is_tuple and has_asdict_method and has_fields_class_member



#!/usr/bin/env python
#
# Copyright (c) 2021 Carsten Igel.
#
# This file is part of nt2dc
# (see https://github.com/carstencodes/nt2dc).
#
# License: 3-clause BSD, see https://opensource.org/licenses/BSD-3-Clause
#

from setuptools import setup, find_packages

__VERSION__ = "0.7.1"

long_description: str = ""
with open("README.md", "r") as read_me_file:
    long_description = read_me_file.read()

setup(
    name="nt2dc",
    version=__VERSION__,
    license="BSD 3-Clause",
    author="Carsten Igel",
    author_email="cig@bite-that-bit.de",
    description="Converter for dataclasses and named tuples",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(where="src"),
    url="https://github.com/carstencodes/nt2dc",
    install_requires=[],
    package_dir={"": "src"},
    keywords="NamedTuple DataClasses",
    python_requires=">=3.8, < 4",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: System",
        "Typing :: Typed",
    ],
)
